"""Keras tools."""
import sklearn.metrics
import numpy as np
import keras
import tensorflow as tf

from . import metrics

class BinaryClassifierMetricsCallback(keras.callbacks.Callback):
    """Keras callback to calculate metrics of a binary classifier for each epoch.

    Attributes
    ----------
    val_data
        The validation data.
    val_labels
        The validation labels.
    pos_label : int, optional
        The positive label number. The default is 1.
    """
    def __init__(self, val_data, val_labels, pos_label=1):
        super().__init__()
        self.val_data = val_data
        self.val_labels = val_labels
        self.pos_label = pos_label

    def on_epoch_end(self, batch, logs={}):
        logs = logs or {}
        predict_results = self.model.predict(self.val_data)

        round_predict_results = np.rint(predict_results)

        roc_auc = sklearn.metrics.roc_auc_score(self.val_labels, predict_results)
        logs["roc_auc"] = roc_auc
        
        average_precision = sklearn.metrics.average_precision_score(self.val_labels, predict_results)
        logs['average_precision'] = average_precision

        f1 = sklearn.metrics.f1_score(self.val_labels, round_predict_results)
        logs["f1"] = f1

        accuracy = sklearn.metrics.accuracy_score(self.val_labels, round_predict_results)
        logs["accuracy"] = accuracy

        best_f1, best_f1_threshold = metrics.best_f1_score(self.val_labels, predict_results, self.pos_label)
        logs["best_f1"] = best_f1
        logs["best_f1_threshold"] = best_f1_threshold


def set_gpu_mem_growth():
    """Allocate only as much GPU memory as needed for Keras.

    Based on runtime allocations: it starts out allocating very little memory,
    and as Sessions get run and more GPU memory is needed, we extend the GPU
    memory region needed by the TensorFlow process. Note that we do not release
    memory, since that can lead to even worse memory fragmentation.
    """
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    keras.backend.tensorflow_backend.set_session(tf.Session(config=config))



"""A collection of plot tools."""
import matplotlib.pyplot as plt

# see https://matplotlib.org/api/_as_gen/matplotlib.axes.Axes.twinx.html
def twin_axes_timeseries_plot(values_1, label_1, values_2, label_2, 
                              start_timestep_number=0, shift_1=0, shift_2=0, 
                              title=None, label_x = 'Step', color_1 = 'tab:red', color_2 = 'tab:blue'):
    """Create twin axes timeseries plot.
    
    Plots two different timeseries curves in one diagram but two different y-axes.

    Parameters
    ----------
    values_1 : array_like
        Values for the first timeseries curve.
    label_1 : str
        Label for the first timeseries curve.
    values_2 : array_like
        Values for the second timeseries curve.
    label_2 : str
        Label for the second timeseries curve.
    start_timestep_number : int, optional
        Number for first point in time. Default is 0.
    shift_1 : int, optional
        Number of timesteps to shift the first timeseries curve.
        Can be positive or negative. Default is 0.
    shift_2 : int, optional
        Number of timesteps to shift the second timeseries curve.
        Can be positive or negative. Default is 0.
    title : str, optional
        Title of the plot.
    label_x : str, optional
        Label for the x-axis (timeseries axis). Default is 'Step'.
    color_1 : str, optional
        Color of first timeseries curve. Default is 'tab:red'.
    color_2 : str, optional
        Color of second timeseries curve. Default is 'tab:blue'.
    """
    fig, ax1 = plt.subplots()
    
    if title != None:
        plt.title(title)

    ax1.set_xlabel(label_x)

    ax1.set_ylabel(label_1, color=color_1)
    ax1.plot(range(start_timestep_number + shift_1, len(values_1) + start_timestep_number + shift_1), 
             values_1, color=color_1)
    ax1.tick_params(axis='y', labelcolor=color_1)

    ax2 = ax1.twinx()
    
    ax2.set_ylabel(label_2, color=color_2)
    ax2.plot(range(start_timestep_number + shift_2, len(values_2) + start_timestep_number + shift_2), 
             values_2, color=color_2)
    ax2.tick_params(axis='y', labelcolor=color_2)

    # otherwise the labels might be slightly clipped
    # see https://matplotlib.org/users/tight_layout_guide.html
    fig.tight_layout() 

    plt.show()

def boxplot(values, labels=None, title=None, xlabel=None, ylabel=None, vert=True):
    """Create boxplot.
    
    Prints one or more boxplots in a single diagram.

    Parameters
    ----------
    values : array_like of numbers for one boxplot or array_like of array_like of numbers for several
        The values to draw the boxplot for. If you want to draw 
        more then one boxplot you have to give an array_like  
        of array_like with numbers.
    labels : str or array_like of str, optional
        The labels of the boxplots.
    title : str, optional
        Title of the plot.    
    xlabel : str, optional
        Label name of the x-axis.
    ylabel : str, optional
        Label name of the y-axis.
    vert : bool, optional
        If True (default), makes the boxes vertical. If False, everything is drawn horizontally.
    """
    _, ax = plt.subplots()
    
    if title is not None:
        ax.set_title(title)

    if xlabel is not None:
        ax.set(xlabel=xlabel)

    if ylabel is not None:
        ax.set(ylabel=ylabel)

    ax.boxplot(values, labels=labels, vert=vert)

    plt.grid(b=True, axis='y', linestyle='--')

    plt.xticks(rotation=90)

    plt.show()

def boxplot_dict(values_dict, title=None, xlabel=None, ylabel=None, vert=True):
    """Create boxplot form dictionary.
    
    Parameters
    ----------
    values_dict : dict with str to array_like
        Dictionary with one entry per box plot. The key (str) 
        is the name of the boxplot, the value (array_like) 
        contains the values to plot.
    title : str, optional
        Title of the plot.    
    xlabel : str, optional
        Label name of the x-axis.
    ylabel : str, optional
        Label name of the y-axis.
    vert : bool, optional
        If True (default), makes the boxes vertical. If False, everything is drawn horizontally.
    """
   
    values = []
    labels = []

    for key, value in values_dict.items():
        values.append(value)
        labels.append(key)
    
    boxplot(values, labels=labels, title=title, xlabel=xlabel, ylabel=ylabel, vert=vert)


"""A collection of generic machine learning tools."""
import itertools
import sys
from tqdm import tqdm
from scipy import stats

def multi_param_call(function, param_dict, iterations, verbose=1):
    """Call function multiple times and return dict with results.

    Calls the given `function` `iterations` times for each entry (value) 
    in the `param_dict`.

    Parameters
    ----------
    function
        The function to call.
    param_dict : dict
        Dictionary with params that are used to call the function.
    iterations : int
        Number of iterations that function will be called for each entry in param_dict.
    verbose : int, optional
        If 1 a progress bar will be shown. If 2 detailed status will be print
        after each iteration. Default is 1.

    Returns
    -------
    dict
        Dict with result values of the `function` calls. If `function` returned just 
        one value the returned dict contains the same keys as `param_dict`. The value is
        an array with one result for each `function` call. The called `function` can also 
        return a dict with results as an alternative. This is if you want to return more 
        then just one result. In this case this result dict contains the same keys
        as the `function` returned on first level. As value it contains a second dict
        as the second level. This second level dict contains the same keys as 
        `param_dict`. The value is an array with one result for each call.
    """
    result = {}
    if verbose == 1:
        number_of_iterations = len(param_dict) * iterations
        pbar = tqdm(total=number_of_iterations, file=sys.stdout)
    for key, value in param_dict.items():
        for i in range(iterations):
            f_result = function(value)

            if isinstance(f_result, dict):
                for f_result_key, f_result_value in f_result.items():
                    sub_dict = result.setdefault(f_result_key, {})
                    sub_dict.setdefault(key, []).append(f_result_value)

            else: 
                result.setdefault(key, []).append(f_result)
            if verbose == 1:
                pbar.update(1)    
            elif verbose == 2:
                print("Done with iteration {} of {} for {}. Result: {}".format(i+1, iterations, key, f_result))
        if verbose == 1:
            pbar.write('Done with {}'.format(key))
    if verbose == 1:
        pbar.close()
    return result

def ttest_combinations(values_dict):
    """Do a t-test on values in a dict and compute the p-value.

    The t-test is computed on each combination of two array_like in the `values_dict`.
    If `values_dict` contains two entries just one p-value is computed. If it contains
    four values this will compute six p-values.

    Parameters
    ----------
    values_dict : dict with key as str and value as array_like
        Dictionary with values to compute the t-test on. At least two entries
        must be present in the dict.

    Returns
    -------
    dict
        Dict with results. Key is a tuple of the compared keys. Value is the p-value.

    See Also
    --------
    Also see the SkiPy function `scipy.stats.ttest_ind`.
    """
    result = {}
    for key_pair in itertools.combinations(values_dict.keys(), 2):
        key_0 = key_pair[0]
        key_1 = key_pair[1]
        result[key_pair] = stats.ttest_ind(values_dict[key_0], values_dict[key_1])[1]
    return result



"""A collection of metrics functions or tools."""
import sklearn

def f1_from_roc(fpr, tpr, pos, neg):
    """Calculate f1 score from roc values.
    
    Parameters
    ----------
        fpr : float
            The false positive rate.
        tpr : float
            The true positive rate.
        pos : int
            The number of positive labels.
        neg : int
            The number of negative labels.
        
    Returns
    -------
    float
        The f1 score.
    """
    fp = fpr * neg
    fn = (1 - tpr) * pos
    tp = pos - fn
    f1 = tp / (tp + ((fn + fp) / 2))
    return f1

def pos_neg(labels, pos_label):
    pos = sum(label == pos_label for label in labels)
    neg = sum(label != pos_label for label in labels)
    return pos, neg        
        
def best_f1_score(labels, predictions, pos_label):
    """Calculate best f1 score with its threshold."""
    fpr, tpr, thresholds = sklearn.metrics.roc_curve(labels, predictions, pos_label=pos_label)
    pos, neg = pos_neg(labels, 1)
    
    best_f1 = -1
    best_f1_threshold = -1
    
    for fpr_value, tpr_value, t in zip(fpr, tpr, thresholds):
        f1 = f1_from_roc(fpr_value, tpr_value, pos, neg)
        if f1 > best_f1:
            best_f1 = f1
            best_f1_threshold = t
    return best_f1, best_f1_threshold        
        



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mltb",
    version="0.0.15",
    maintainer="Philip May",
    author="Philip May",
    author_email="pm@eniak.de",
    description="Machine Learning Tool Box",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PhilipMay/mltb/tree/master/mltb",
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=[
        'sklearn', 
        'numpy', 
        'keras',
        'matplotlib',
        'tqdm',
    ],
    keywords='keras metric',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
    ],
)